from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from typing import List, IO


class PeriodoImpressaoCortesEstados(Section):
    """
    Bloco do arquivo nwlistcf.dat que armazena as informações de mês de
    início e fim para a impressão.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [IntegerField(3, 1), IntegerField(3, 5), IntegerField(1, 9)]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PeriodoImpressaoCortesEstados):
            return False
        bloco: PeriodoImpressaoCortesEstados = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO, *args, **kwargs):
        # Lê as linhas de cabeçalho
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.data = self.__linha.read(file.readline())

    # Override
    def write(self, file: IO, *args, **kwargs):
        # Escreve as linhas de cabeçalho
        for c in self.__cabecalhos:
            file.write(c)

        file.write(self.__linha.write(self.data))


class OpcoesImpressao(Section):
    """
    Bloco do arquivo nwlistcf.dat que armazena as informações das opções
    de impressão selecionadas.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [IntegerField(2, 1), IntegerField(2, 4), IntegerField(2, 7)]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, OpcoesImpressao):
            return False
        bloco: OpcoesImpressao = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO, *args, **kwargs):
        # Lê as linhas de cabeçalho
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.data = self.__linha.read(file.readline())

    # Override
    def write(self, file: IO, *args, **kwargs):
        # Escreve as linhas de cabeçalho
        for c in self.__cabecalhos:
            file.write(c)

        file.write(self.__linha.write(self.data))



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.datetimefield import DatetimeField
from typing import List, IO, Optional
import pandas as pd  # type: ignore
from datetime import datetime


class BlocoUHEGhmin(Section):
    """
    Bloco com os valores de geração hidráulica mínima.
    """

    FIM_BLOCO = "999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_uhe = Line(
            [
                IntegerField(3, 0),
                DatetimeField(7, 5, format="%m %Y"),
                IntegerField(1, 14),
                FloatField(6, 17, 0),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoUHEGhmin):
            return False
        bloco: BlocoUHEGhmin = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame()
            df["codigo_usina"] = codigos
            df["data"] = datas
            df["patamar"] = patamares
            df["geracao"] = geracoes
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Variáveis auxiliares
        codigos: List[Optional[int]] = []
        datas: List[Optional[datetime]] = []
        patamares: List[Optional[int]] = []
        geracoes: List[Optional[float]] = []

        while True:
            linha = file.readline()
            # Confere se acabou
            if BlocoUHEGhmin.FIM_BLOCO in linha[:3] or len(linha) < 3:
                if len(codigos) > 0:
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha_uhe.read(linha)
            codigos.append(dados[0])
            datas.append(dados[1])
            patamares.append(dados[2])
            geracoes.append(dados[3])

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do ghmin.dat não foram lidos com sucesso")

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            dados_linha = linha_lida.tolist()
            file.write(self.__linha_uhe.write(dados_linha))
        file.write(f"{BlocoUHEGhmin.FIM_BLOCO}\n")



from inewave.nwlistop.nwlistopdat import Nwlistopdat

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.nwlistopdat import (
    MockNwlistopdatOp1,
    MockNwlistopdatOp2,
    MockNwlistopdatOp4,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_nwlistopdat_op1():
    m: MagicMock = mock_open(read_data="".join(MockNwlistopdatOp1))
    with patch("builtins.open", m):
        n = Nwlistopdat.read(ARQ_TESTE)
        assert n.periodo_inicial_impressao == 1
        assert n.periodo_final_impressao == 1
        assert n.serie_inicial_impressao == 1
        assert n.serie_final_impressao == 1


def test_eq_nwlistopdat_op1():
    m: MagicMock = mock_open(read_data="".join(MockNwlistopdatOp1))
    with patch("builtins.open", m):
        n1 = Nwlistopdat.read(ARQ_TESTE)
        n2 = Nwlistopdat.read(ARQ_TESTE)
        assert n1 == n2


def test_neq_nwlistopdat_op1():
    m: MagicMock = mock_open(read_data="".join(MockNwlistopdatOp1))
    with patch("builtins.open", m):
        n1 = Nwlistopdat.read(ARQ_TESTE)
        n2 = Nwlistopdat.read(ARQ_TESTE)
        n1.periodo_inicial_impressao = 49
        assert n1 != n2
        n1.periodo_inicial_impressao = 1
        n1.periodo_final_impressao = 49
        assert n1 != n2
        n1.periodo_final_impressao = 1
        n1.serie_inicial_impressao = 10
        assert n1 != n2
        n1.serie_inicial_impressao = 1
        n1.serie_final_impressao = 20
        assert n1 != n2


def test_atributos_encontrados_nwlistopdat_op2():
    m: MagicMock = mock_open(read_data="".join(MockNwlistopdatOp2))
    with patch("builtins.open", m):
        n = Nwlistopdat.read(ARQ_TESTE)
        assert n.periodo_inicial_impressao == 1
        assert n.periodo_final_impressao == 58
        assert n.variaveis_impressao_estagios_agregados[0] == 99
        assert n.variaveis_impressao_estagios_individualizados[0] == 99
        assert n.uhes_impressao_estagios_individualizados[0] == 999


def test_eq_nwlistopdat_op2():
    m: MagicMock = mock_open(read_data="".join(MockNwlistopdatOp2))
    with patch("builtins.open", m):
        n1 = Nwlistopdat.read(ARQ_TESTE)
        n2 = Nwlistopdat.read(ARQ_TESTE)
        assert n1 == n2


def test_neq_nwlistopdat_op2():
    m: MagicMock = mock_open(read_data="".join(MockNwlistopdatOp2))
    with patch("builtins.open", m):
        n1 = Nwlistopdat.read(ARQ_TESTE)
        n2 = Nwlistopdat.read(ARQ_TESTE)
        n2.variaveis_impressao_estagios_agregados[0] = 1
        assert n1 != n2


def test_eq_nwlistopdat_op4():
    m: MagicMock = mock_open(read_data="".join(MockNwlistopdatOp4))
    with patch("builtins.open", m):
        n1 = Nwlistopdat.read(ARQ_TESTE)
        n2 = Nwlistopdat.read(ARQ_TESTE)
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from inewave.config import (
    MAX_AGRUPAMENTOS_INTERCAMBIOS,
    MAX_SUBMERCADOS,
    MAX_MESES_ESTUDO,
)

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.datetimefield import DatetimeField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
from datetime import datetime
import pandas as pd  # type: ignore
import numpy as np  # type: ignore
from inewave._utils.formatacao import (
    repete_vetor,
)


class BlocoGruposAgrint(Section):
    """
    Bloco com informações dos intercâmbios pertencentes aos
    grupos.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 1),
                IntegerField(3, 5),
                IntegerField(3, 9),
                FloatField(7, 13, 4),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoGruposAgrint):
            return False
        bloco: BlocoGruposAgrint = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            cols = [
                "agrupamento",
                "submercado_de",
                "submercado_para",
                "coeficiente",
            ]
            df = pd.DataFrame(tabela, columns=cols)
            df = df.astype(
                {
                    "agrupamento": "int64",
                    "submercado_de": "int64",
                    "submercado_para": "int64",
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_AGRUPAMENTOS_INTERCAMBIOS * MAX_SUBMERCADOS, 4))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoGruposAgrint.FIM_BLOCO in linha[:4]:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                tabela[i, :] = dados
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do agrint.dat não foram lidos com sucesso")

        for _, linha in self.data.iterrows():
            linha_escrita = []
            for v in linha:
                linha_escrita.append(None if np.isnan(v) else int(v))
            file.write(self.__linha.write(linha_escrita))
        file.write(BlocoGruposAgrint.FIM_BLOCO + "\n")


class BlocoLimitesPorGrupoAgrint(Section):
    """
    Bloco com informações de configuração dos limites dos agrupamentos
    de intercâmbio por período de estudo.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 1),
                DatetimeField(7, 6, format="%m %Y"),
                DatetimeField(7, 14, format="%m %Y"),
                FloatField(7, 22, 0),
                FloatField(7, 30, 0),
                FloatField(7, 38, 0),
                LiteralField(40, 50),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoLimitesPorGrupoAgrint):
            return False
        bloco: BlocoLimitesPorGrupoAgrint = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "agrupamento": repete_vetor(agrupamentos, 3),
                    "data_inicio": repete_vetor(datas_inicio, 3),
                    "data_fim": repete_vetor(datas_fim, 3),
                    "comentario": repete_vetor(comentarios, 3),
                    "patamar": np.tile(
                        np.arange(1, 3 + 1),
                        len(agrupamentos),
                    ),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        agrupamentos: List[int] = []
        datas_inicio: List[datetime] = []
        datas_fim: List[datetime] = []
        comentarios: List[str] = []
        tabela = np.zeros(
            (MAX_AGRUPAMENTOS_INTERCAMBIOS * MAX_MESES_ESTUDO, 3)
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoGruposAgrint.FIM_BLOCO in linha[:4]:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                agrupamentos.append(dados[0])
                tabela[i, :] = dados[3:6]
                datas_inicio.append(dados[1])
                datas_fim.append(dados[2])
                comentarios.append(dados[-1])
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do agrint.dat não foram lidos com sucesso")

        df = self.data.copy()
        for _, linha_agrupamento in (
            self.data[["agrupamento", "data_inicio", "data_fim", "comentario"]]
            .drop_duplicates()
            .iterrows()
        ):
            df_agrupamento = df.loc[
                (df["agrupamento"] == linha_agrupamento["agrupamento"])
                & (df["data_inicio"] == linha_agrupamento["data_inicio"])
            ]
            if not pd.isna(linha_agrupamento["data_fim"]):
                df_agrupamento = df_agrupamento.loc[
                    (
                        df_agrupamento["data_fim"]
                        == linha_agrupamento["data_fim"]
                    )
                ]
            file.write(
                self.__linha.write(
                    [
                        linha_agrupamento["agrupamento"],
                        linha_agrupamento["data_inicio"],
                        linha_agrupamento["data_fim"],
                    ]
                    + df_agrupamento.sort_values("patamar")["valor"].tolist()
                    + [linha_agrupamento["comentario"]]
                )
            )
        file.write(BlocoGruposAgrint.FIM_BLOCO + "\n")



from inewave.newave.modelos.shist import (
    BlocoVarreduraShist,
    BlocoSeriesSimulacaoShist,
)

from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional


class Shist(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à varredura
    das séries históricas para simulação.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [
        BlocoVarreduraShist,
        BlocoSeriesSimulacaoShist,
    ]

    @property
    def varredura(self) -> Optional[int]:
        """
        Flag para habilitar o uso de varredura na simulação.

        :return: O valor do flag
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoVarreduraShist)
        if isinstance(b, BlocoVarreduraShist):
            return b.data[0]
        return None

    @varredura.setter
    def varredura(self, valor: int):
        b = self.data.get_sections_of_type(BlocoVarreduraShist)
        if isinstance(b, BlocoVarreduraShist):
            b.data[0] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def ano_inicio_varredura(self) -> Optional[int]:
        """
        O ano de início da varredura na simulação.

        :return: O ano
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoVarreduraShist)
        if isinstance(b, BlocoVarreduraShist):
            return b.data[1]
        return None

    @ano_inicio_varredura.setter
    def ano_inicio_varredura(self, valor: int):
        b = self.data.get_sections_of_type(BlocoVarreduraShist)
        if isinstance(b, BlocoVarreduraShist):
            b.data[1] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def anos_inicio_simulacoes(self) -> Optional[List[int]]:
        """
        Os anos de início das simulações.

        :return: Os anos
        :rtype: list[int] | None
        """
        b = self.data.get_sections_of_type(BlocoSeriesSimulacaoShist)
        if isinstance(b, BlocoSeriesSimulacaoShist):
            return b.data
        return None

    @anos_inicio_simulacoes.setter
    def anos_inicio_simulacoes(self, valor: List[int]):
        b = self.data.get_sections_of_type(BlocoSeriesSimulacaoShist)
        if isinstance(b, BlocoSeriesSimulacaoShist):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.restricaoeletrica import (
    RegistroRE,
    RegistroREHorizPer,
    RegistroRELimFormPer,
)


class RestricaoEletrica(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao cadastro
    das restrições lineares por partes no domínio de energia (REE).
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroRELimFormPer,
        RegistroREHorizPer,
        RegistroRE,
    ]

    def re(
        self,
        codigo_restricao: Optional[int] = None,
        formula: Optional[str] = None,
    ) -> Optional[Union[RegistroRE, List[RegistroRE]]]:
        """
        Obtém um registro que cadastra uma usina restrição elétrica (RE).

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param formula: equação da restrição
        :type formula: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRE` |
            list[:class:`RegistroRE`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRE,
            codigo_restricao=codigo_restricao,
            formula=formula,
        )

    def re_horiz_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
    ) -> Optional[Union[RegistroREHorizPer, List[RegistroREHorizPer]]]:
        """
        Obtém um registro que cadastra o horizonte de validade de uma
        restrição elétrica.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicial: data inicial de validade da restrição
        :type data_inicial: datetime | None
        :param data_final: data final de validade da restrição
        :type data_final: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroREHorizPer` |
            list[:class:`RegistroREHorizPer`] | None
        """
        return self.data.get_registers_of_type(
            RegistroREHorizPer,
            codigo_restricao=codigo_restricao,
            data_inicial=data_inicial,
            data_final=data_final,
        )

    def re_lim_form_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicial: Optional[datetime] = None,
        data_final: Optional[datetime] = None,
        patamar: Optional[int] = None,
        limite_inferior: Optional[float] = None,
        limite_superior: Optional[float] = None,
    ) -> Optional[Union[RegistroRELimFormPer, List[RegistroRELimFormPer]]]:
        """
        Obtém um registro que cadastra os limites por horizonte e por
        patamar para uma restrição elétrica.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicial: data inicial de validade dos limites
        :type data_inicial: datetime | None
        :param data_final: data final de validade dos limites
        :type data_final: datetime | None
        :param patamar: patamar de validade dos limites
        :type patamar: int | None
        :param limite_inferior: limite inferior da restrição
        :type limite_inferior: float | None
        :param limite_superior: limite superior da restrição
        :type limite_superior: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRELimFormPer` |
            list[:class:`RegistroRELimFormPer`] | None
        """
        return self.data.get_registers_of_type(
            RegistroRELimFormPer,
            codigo_restricao=codigo_restricao,
            data_inicial=data_inicial,
            data_final=data_final,
            patamar=patamar,
            limite_inferior=limite_inferior,
            limite_superior=limite_superior,
        )



from inewave.nwlistcf.modelos.caso import NomeCaso

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional


class Caso(SectionFile):
    """
    Armazena os dados de entrada do NWLISTCF referentes ao caso de estudo.

    Esta classe lida com informações de entrada fornecidas ao NWLISTCF e
    que podem ser modificadas através do arquivo `caso.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [NomeCaso]

    @property
    def arquivos(self) -> Optional[str]:
        """
        Caminho para o arquivo `arquivos.dat` de entrada do NWLISTCF.

        :return: O caminho para o arquivo
        :rtype: str | None
        """
        b = self.data.get_sections_of_type(NomeCaso)
        if isinstance(b, NomeCaso):
            return b.data
        return None

    @arquivos.setter
    def arquivos(self, a: str):
        b = self.data.get_sections_of_type(NomeCaso)
        if isinstance(b, NomeCaso):
            b.data = a



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.blocos.valoresseriepatamar import (
    ValoresSeriePatamar,
)

from cfinterface.files.blockfile import BlockFile
import pandas as pd  # type: ignore
from typing import TypeVar, Optional


class ArquivoSubmercadoPatamar(BlockFile):
    """
    Armazena os dados das saídas por patamar, por submercado.
    """

    T = TypeVar("T")

    BLOCKS = [Submercado, ValoresSeriePatamar]

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__valores = None

    def __monta_tabela(self) -> pd.DataFrame:
        df = None
        for b in self.data.of_type(ValoresSeriePatamar):
            dados = b.data
            if dados is None:
                continue
            elif df is None:
                df = b.data
            else:
                df = pd.concat([df, b.data], ignore_index=True)
        return df

    @property
    def valores(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os valores por patamar, por série e
        por mês/ano de estudo.

        - data (`datetime`)
        - patamar (`str`)
        - serie (`str`)
        - valor (`float`)

        :return: A tabela dos valores por patamar.
        :rtype: pd.DataFrame | None
        """
        if self.__valores is None:
            self.__valores = self.__monta_tabela()
        return self.__valores

    @property
    def submercado(self) -> Optional[str]:
        """
        O submercado associado ao arquivo lido.

        :return: Os nome do submercado
        :rtype: str
        """
        b = self.data.get_blocks_of_type(Submercado)
        if isinstance(b, Submercado):
            return b.data
        return None



from inewave.newave.modelos.clast import (
    BlocoUTEClasT,
    BlocoModificacaoUTEClasT,
)

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Clast(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às classes de
    usinas térmicas.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoUTEClasT, BlocoModificacaoUTEClasT]

    @property
    def usinas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as usinas e seus custos.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - tipo_combustivel (`str`)
        - indice_ano_estudo (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoUTEClasT)
        if isinstance(b, BlocoUTEClasT):
            return b.data
        return None

    @usinas.setter
    def usinas(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoUTEClasT)
        if isinstance(b, BlocoUTEClasT):
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def modificacoes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as modificações de custos das usinas
        organizadas por usina.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - data_inicio (`datetime`)
        - data_fim (`datetime`)
        - custo (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoModificacaoUTEClasT)
        if isinstance(b, BlocoModificacaoUTEClasT):
            return b.data
        return None

    @modificacoes.setter
    def modificacoes(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoModificacaoUTEClasT)
        if isinstance(b, BlocoModificacaoUTEClasT):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from cfinterface.components.section import Section
from typing import IO, List
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class SecaoDadosCortes(Section):
    """
    Registro com os cortes da FCF.
    """

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, SecaoDadosCortes):
            return False
        bloco: SecaoDadosCortes = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(bloco.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    def __le_e_atribui_int(
        self, file: IO, destino: np.ndarray, tamanho: int, indice: int
    ):
        destino[indice, :] = np.frombuffer(
            file.read(tamanho * 4),
            dtype=np.int32,
            count=tamanho,
        )

    def __le_e_atribui_float(
        self, file: IO, destino: np.ndarray, tamanho: int, indice: int
    ):
        destino[indice, :] = np.frombuffer(
            file.read(tamanho * 8),
            dtype=np.float64,
            count=tamanho,
        )

    def __inicializa_variaveis(self):
        self.__tabela_int = np.zeros(
            (self.__numero_total_cortes, 4), dtype=np.int32
        )
        bytes_nao_coeficientes = 4 * 4
        self.__numero_coeficientes = int(
            (self.__tamanho_registro - bytes_nao_coeficientes) / 8
        )
        self.__tabela_float = np.zeros(
            (self.__numero_total_cortes, self.__numero_coeficientes),
            dtype=np.float64,
        )

    def __le_registro(
        self,
        file: IO,
        offset: int,
        indice: int,
    ) -> int:
        file.seek(offset)
        self.__le_e_atribui_int(file, self.__tabela_int, 4, indice)
        self.__le_e_atribui_float(
            file, self.__tabela_float, self.__numero_coeficientes, indice
        )
        indice_proximo_corte = self.__tabela_int[indice, 0]
        self.__tabela_int[indice, 0] = indice
        return indice_proximo_corte

    def __converte_array_ree(self, cortes_lidos: int) -> pd.DataFrame:
        self.__tabela_int = self.__tabela_int[:cortes_lidos, :]
        df_int = pd.DataFrame(
            self.__tabela_int,
            columns=[
                "indice_corte",
                "iteracao_construcao",
                "indice_forward",
                "iteracao_desativacao",
            ],
        )
        # O índice guardado é o índice do corte anterior.
        # Para transformar no índice do corte em si, soma 1
        df_int["indice_corte"] += 1
        self.__tabela_float = self.__tabela_float[:cortes_lidos, :]

        cols_earm = [f"pi_earm_ree{i}" for i in self.__codigos_rees]
        cols_ena = [
            f"pi_ena_ree{i}_lag{j}"
            for i in self.__codigos_rees
            for j in range(1, self.__ordem_maxima_parp + 1)
        ]
        cols_gnl = [
            f"pi_gnl_sbm{i}_pat{j}_lag{k}"
            for i in self.__codigos_submercados
            for j in range(1, self.__numero_patamares_carga + 1)
            for k in range(1, self.__lag_maximo_gnl + 1)
        ]

        num_cols = 1 + len(cols_earm) + len(cols_ena) + len(cols_gnl)
        df_float = pd.DataFrame(
            self.__tabela_float[:, :num_cols],
            columns=["rhs"] + cols_earm + cols_ena + cols_gnl,
        )
        return pd.concat([df_int, df_float], axis=1)

    def __converte_array_indiv(self, cortes_lidos: int) -> pd.DataFrame:
        self.__tabela_int = self.__tabela_int[:cortes_lidos, :]
        df_int = pd.DataFrame(
            self.__tabela_int,
            columns=[
                "indice_corte",
                "iteracao_construcao",
                "indice_forward",
                "iteracao_desativacao",
            ],
        )
        # O índice guardado é o índice do corte anterior.
        # Para transformar no índice do corte em si, soma 1
        df_int["indice_corte"] += 1
        self.__tabela_float = self.__tabela_float[:cortes_lidos, :]

        cols_varm = [f"pi_varm_uhe{i}" for i in self.__codigos_uhes]
        cols_qafl = [
            f"pi_qafl_uhe{i}_lag{j}"
            for i in self.__codigos_uhes
            for j in range(1, self.__ordem_maxima_parp + 1)
        ]
        cols_gnl = [
            f"pi_gnl_sbm{i}_pat{j}_lag{k}"
            for i in self.__codigos_submercados
            for j in range(1, self.__numero_patamares_carga + 1)
            for k in range(1, self.__lag_maximo_gnl + 1)
        ]

        num_cols_gnl = len(cols_gnl)
        num_cols_varm_qafl = len(self.__codigos_uhes) * (
            self.__ordem_maxima_parp + 1
        )
        offset_inicio_cols_varm = num_cols_gnl + 2
        indices_cols_df_float = (
            [0]
            + list(range(1, num_cols_gnl + 1))
            + list(
                range(
                    offset_inicio_cols_varm,
                    offset_inicio_cols_varm + num_cols_varm_qafl,
                )
            )
        )

        df_float = pd.DataFrame(
            self.__tabela_float[:, indices_cols_df_float],
            columns=["rhs"] + cols_gnl + cols_varm + cols_qafl,
        )
        return pd.concat([df_int, df_float], axis=1)

    def __converte_array_em_dataframe(self, cortes_lidos: int):
        if len(self.__codigos_rees) > 0:
            df = self.__converte_array_ree(cortes_lidos)
        elif len(self.__codigos_uhes) > 0:
            df = self.__converte_array_indiv(cortes_lidos)
        else:
            df = pd.DataFrame(
                columns=[
                    "indice_corte",
                    "iteracao_construcao",
                    "indice_forward",
                    "iteracao_desativacao",
                ],
            )
        # Inverte a indexação
        df["indice_corte"] = df["indice_corte"].to_numpy()[::-1]
        self.data = df

    def read(
        self,
        file: IO,
        tamanho_registro: int = 1664,
        indice_ultimo_corte: int = 1,
        numero_total_cortes: int = 10000,
        codigos_rees: List[str] = [],
        codigos_uhes: List[str] = [],
        codigos_submercados: List[str] = [],
        ordem_maxima_parp: int = 12,
        numero_patamares_carga: int = 3,
        lag_maximo_gnl: int = 2,
        *args,
        **kwargs,
    ):
        # Atribui variáveis locais
        self.__tamanho_registro = tamanho_registro
        self.__numero_total_cortes = numero_total_cortes
        self.__codigos_rees = codigos_rees
        self.__codigos_uhes = codigos_uhes
        self.__codigos_submercados = codigos_submercados
        self.__ordem_maxima_parp = ordem_maxima_parp
        self.__numero_patamares_carga = numero_patamares_carga
        self.__lag_maximo_gnl = lag_maximo_gnl
        # Realiza leitura
        self.__inicializa_variaveis()
        indice_proximo_corte = self.__le_registro(
            file, (indice_ultimo_corte - 1) * tamanho_registro, 0
        )
        cortes_lidos = 1
        while (indice_proximo_corte != 0) and (
            cortes_lidos < numero_total_cortes
        ):
            indice_proximo_corte = self.__le_registro(
                file,
                tamanho_registro * (indice_proximo_corte - 1),
                cortes_lidos,
            )
            cortes_lidos += 1
        self.__converte_array_em_dataframe(cortes_lidos)



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)

from inewave.nwlistop.modelos.evaporsin import EvapoAnos


class Evaporsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes à energia
    de evaporação para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `evaporsin.out`.

    """

    BLOCKS = [
        EvapoAnos,
    ]



# Imports de módulos externos
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField

from inewave.newave.modelos.blocos.tabelacsv import TabelaCSV


class TabelaAvlCortesFpha(TabelaCSV):
    """
    Bloco com as informações dos cortes da função de produção para as
    UHEs do modelo NEWAVE.
    """

    BEGIN_PATTERN = "-----;-----;--------------;"
    LINE_MODEL = Line(
        [
            IntegerField(size=5),
            IntegerField(size=5),
            LiteralField(size=14),
            IntegerField(size=7),
            FloatField(size=10, decimal_digits=6),
            FloatField(size=16, decimal_digits=8),
            FloatField(size=16, decimal_digits=8),
            FloatField(size=16, decimal_digits=8),
            FloatField(size=16, decimal_digits=8),
            FloatField(size=16, decimal_digits=8),
        ],
        delimiter=";",
    )
    COLUMN_NAMES = [
        "codigo_usina",
        "periodo",
        "nome_usina",
        "indice_corte",
        "fator_correcao",
        "rhs_energia",
        "coeficiente_volume_util_MW_hm3",
        "coeficiente_vazao_turbinada_MW_m3s",
        "coeficiente_vazao_vertida_MW_m3s",
        "coeficiente_vazao_lateral_MW_m3s",
    ]
    END_PATTERN = ""



from inewave.nwlistcf.modelos.arquivos import BlocoNomesArquivos
from inewave.nwlistcf import Arquivos

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.arquivos_nwlistcf import MockArquivos

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_nomes_arquivos():
    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    b = BlocoNomesArquivos()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 15
    assert b.data.shape[1] == 2
    assert b.data.iloc[0, 1] == "nwlistcf.dat"
    assert b.data.iloc[-1, 1] == "vazaoxf.dat"


def test_atributos_encontrados_arquivos():
    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m):
        ad = Arquivos.read(ARQ_TESTE)
        assert ad.nwlistcf is not None
        assert ad.cortes is not None
        assert ad.cortesh is not None
        assert ad.newdesp is not None
        assert ad.cortese is not None
        assert ad.energiaf is not None
        assert ad.rsar is not None
        assert ad.rsarh is not None
        assert ad.rsari is not None
        assert ad.relatorio_cortes is not None
        assert ad.relatorio_estados is not None
        assert ad.relatorio_rsar is not None
        assert ad.energiaxf is not None
        assert ad.vazaof is not None
        assert ad.vazaoxf is not None


def test_atributos_nao_encontrados_arquivos():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Arquivos.read(ARQ_TESTE)
        assert ad.nwlistcf is None
        assert ad.cortes is None
        assert ad.cortesh is None
        assert ad.newdesp is None
        assert ad.cortese is None
        assert ad.energiaf is None
        assert ad.rsar is None
        assert ad.rsarh is None
        assert ad.rsari is None
        assert ad.relatorio_cortes is None
        assert ad.relatorio_estados is None
        assert ad.relatorio_rsar is None
        assert ad.energiaxf is None
        assert ad.vazaof is None
        assert ad.vazaoxf is None


def test_eq_arquivos():
    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m):
        ad1 = Arquivos.read(ARQ_TESTE)
        ad2 = Arquivos.read(ARQ_TESTE)
        assert ad1 == ad2


def test_neq_arquivos():
    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m):
        ad1 = Arquivos.read(ARQ_TESTE)
        ad2 = Arquivos.read(ARQ_TESTE)
        ad2.nwlistcf = "teste"
        assert ad1 != ad2


def test_leitura_escrita_arquivos():
    m_leitura: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m_leitura):
        ad1 = Arquivos.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ad1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ad2 = Arquivos.read(ARQ_TESTE)
        assert ad1 == ad2



from inewave.newave.modelos.agrint import (
    BlocoGruposAgrint,
    BlocoLimitesPorGrupoAgrint,
)

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Agrint(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos agrupamentos
    de intercâmbio.

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoGruposAgrint,
        BlocoLimitesPorGrupoAgrint,
    ]

    @property
    def agrupamentos(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os intercâmbios em cada agrupamento.

        - agrupamento (`int`)
        - submercado_de (`int`)
        - submercado_para (`int`)
        - coeficiente (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoGruposAgrint)
        if isinstance(b, BlocoGruposAgrint):
            return b.data
        return None

    @agrupamentos.setter
    def agrupamentos(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoGruposAgrint)
        if isinstance(b, BlocoGruposAgrint):
            b.data = df

    @property
    def limites_agrupamentos(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os limites dos agrupamentos de intercâmbio
        durante o período de estudo.

        - agrupamento (`int`)
        - data_inicio (`datetime`)
        - data_fim (`datetime`)
        - comentario (`str`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoLimitesPorGrupoAgrint)
        if isinstance(b, BlocoLimitesPorGrupoAgrint):
            return b.data
        return None

    @limites_agrupamentos.setter
    def limites_agrupamentos(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoLimitesPorGrupoAgrint)
        if isinstance(b, BlocoLimitesPorGrupoAgrint):
            b.data = df



# Imports de módulos externos
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField

from inewave.newave.modelos.blocos.tabelacsv import TabelaCSV


class TabelaEcoFpha(TabelaCSV):
    """
    Bloco com as informações da função de produção para as
    UHEs do modelo NEWAVE.
    """

    BEGIN_PATTERN = "----;----;-------------;----;"
    LINE_MODEL = Line(
        [
            IntegerField(size=4),
            IntegerField(size=4),
            LiteralField(size=13),
            IntegerField(size=4),
            IntegerField(size=4),
            IntegerField(size=4),
            IntegerField(size=4),
            IntegerField(size=5),
            FloatField(size=8, decimal_digits=1),
            FloatField(size=8, decimal_digits=1),
            IntegerField(size=5),
            FloatField(size=8, decimal_digits=1),
            FloatField(size=8, decimal_digits=1),
            FloatField(size=8, decimal_digits=1),
            FloatField(size=8, decimal_digits=1),
        ],
        delimiter=";",
    )
    COLUMN_NAMES = [
        "codigo_usina",
        "periodo",
        "nome_usina",
        "tipo",
        "conv",
        "alfa",
        "rems",
        "numero_pontos_vazao_turbinada",
        "vazao_turbinada_minima",
        "vazao_turbinada_maxima",
        "numero_pontos_volume_armazenado",
        "volume_armazenado_minimo",
        "volume_armazenado_maximo",
        "geracao_minima",
        "geracao_maxima",
    ]
    END_PATTERN = ""



from cfinterface.components.section import Section
import numpy as np  # type: ignore
from typing import List
from typing import IO


class SecaoDadosForwarh(Section):
    """
    Registro com os dados da execução do caso existente no
    arquivo forwarh.dat
    """

    ZVAZ = 100

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, SecaoDadosForwarh):
            return False
        bloco: SecaoDadosForwarh = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    def read(self, file: IO, *args, **kwargs):
        _ = np.frombuffer(
            file.read(1 * 4),
            dtype=np.int32,
            count=1,
        )
        nome_caso = file.read(80).decode("utf-8").strip()
        dados_primeiro_bloco = np.frombuffer(
            file.read(7 * 4),
            dtype=np.int32,
            count=7,
        )
        self.data = [nome_caso] + list(dados_primeiro_bloco)
        dados_segundo_bloco = np.frombuffer(
            file.read((self.numero_submercados + 12) * 4),
            dtype=np.int32,
            count=self.numero_submercados + 12,
        )
        self.data += list(dados_segundo_bloco)
        dados_terceiro_bloco = np.frombuffer(
            file.read((self.__class__.ZVAZ + 4) * 4),
            dtype=np.int32,
            count=self.__class__.ZVAZ + 4,
        )
        self.data += list(dados_terceiro_bloco)

    def __offset_primeiro_bloco(self):
        return 7 + self.numero_submercados

    def __offset_segundo_bloco(self):
        return self.__offset_primeiro_bloco() + 13 + self.__class__.ZVAZ

    @property
    def nome_caso(self) -> str:
        return self.data[0]

    @nome_caso.setter
    def nome_caso(self, v: str):
        self.data[0] = v

    @property
    def numero_rees(self) -> int:
        return self.data[1]

    @numero_rees.setter
    def numero_rees(self, v: int):
        self.data[1] = v

    @property
    def numero_submercados(self) -> int:
        return self.data[2]

    @numero_submercados.setter
    def numero_submercados(self, v: int):
        self.data[2] = v

    @property
    def numero_series_gravadas(self) -> int:
        return self.data[3]

    @numero_series_gravadas.setter
    def numero_series_gravadas(self, v: int):
        self.data[3] = v

    @property
    def numero_aberturas(self) -> int:
        return self.data[4]

    @numero_aberturas.setter
    def numero_aberturas(self, v: int):
        self.data[4] = v

    @property
    def numero_estagios_estudo(self) -> int:
        return self.data[5]

    @numero_estagios_estudo.setter
    def numero_estagios_estudo(self, v: int):
        self.data[5] = v

    @property
    def intervalo_series_gravadas(self) -> int:
        return self.data[6]

    @intervalo_series_gravadas.setter
    def intervalo_series_gravadas(self, v: int):
        self.data[6] = v

    @property
    def numero_classes_termicas_submercados(self) -> List[int]:
        return self.data[7 : self.__offset_primeiro_bloco()]

    @numero_classes_termicas_submercados.setter
    def numero_classes_termicas_submercados(self, v: List[int]):
        self.data = (
            self.data[:7] + v + self.data[: self.__offset_primeiro_bloco()]
        )

    @property
    def numero_patamares_deficit(self) -> int:
        return self.data[self.__offset_primeiro_bloco()]

    @numero_patamares_deficit.setter
    def numero_patamares_deficit(self, v: int):
        self.data[self.__offset_primeiro_bloco()] = v

    @property
    def tamanho_registro_arquivo_forward(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 1]

    @tamanho_registro_arquivo_forward.setter
    def tamanho_registro_arquivo_forward(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 1] = v

    @property
    def numero_registros_arquivo_forward(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 2]

    @numero_registros_arquivo_forward.setter
    def numero_registros_arquivo_forward(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 2] = v

    @property
    def numero_registros_necessarios_estagio(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 3]

    @numero_registros_necessarios_estagio.setter
    def numero_registros_necessarios_estagio(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 3] = v

    @property
    def ano_inicio_estudo(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 4]

    @ano_inicio_estudo.setter
    def ano_inicio_estudo(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 4] = v

    @property
    def ano_inicio_historico_vazoes(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 5]

    @ano_inicio_historico_vazoes.setter
    def ano_inicio_historico_vazoes(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 5] = v

    @property
    def numero_anos_descontar_historico_vazoes(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 6]

    @numero_anos_descontar_historico_vazoes.setter
    def numero_anos_descontar_historico_vazoes(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 6] = v

    @property
    def numero_estagios_ano(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 7]

    @numero_estagios_ano.setter
    def numero_estagios_ano(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 7] = v

    @property
    def mes_inicio_estudo(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 8]

    @mes_inicio_estudo.setter
    def mes_inicio_estudo(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 8] = v

    @property
    def mes_inicio_pre_estudo(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 9]

    @mes_inicio_pre_estudo.setter
    def mes_inicio_pre_estudo(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 9] = v

    @property
    def numero_estagios_pre_estudo(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 10]

    @numero_estagios_pre_estudo.setter
    def numero_estagios_pre_estudo(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 10] = v

    @property
    def numero_patamares_carga(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 11]

    @numero_patamares_carga.setter
    def numero_patamares_carga(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 11] = v

    @property
    def ordem_maxima_parp(self) -> int:
        return self.data[self.__offset_primeiro_bloco() + 12]

    @ordem_maxima_parp.setter
    def ordem_maxima_parp(self, v: int):
        self.data[self.__offset_primeiro_bloco() + 12] = v

    @property
    def ano_inicio_series_historicas_simuladas(self) -> List[int]:
        return self.data[
            self.__offset_primeiro_bloco() + 13 : self.__offset_segundo_bloco()
        ]

    @ano_inicio_series_historicas_simuladas.setter
    def ano_inicio_series_historicas_simuladas(self, v: List[int]):
        self.data = (
            self.data[self.__offset_primeiro_bloco() + 13 :]
            + v
            + self.data[self.__offset_segundo_bloco() :]
        )

    @property
    def numero_anos_historico_vazoes(self) -> int:
        return self.data[self.__offset_segundo_bloco()]

    @numero_anos_historico_vazoes.setter
    def numero_anos_historico_vazoes(self, v: int):
        self.data[self.__offset_segundo_bloco()] = v

    @property
    def numero_total_submercados(self) -> int:
        return self.data[self.__offset_segundo_bloco() + 1]

    @numero_total_submercados.setter
    def numero_total_submercados(self, v: int):
        self.data[self.__offset_segundo_bloco() + 1] = v

    @property
    def simulacao_final_individualizada(self) -> int:
        return self.data[self.__offset_segundo_bloco() + 2]

    @simulacao_final_individualizada.setter
    def simulacao_final_individualizada(self, v: int):
        self.data[self.__offset_segundo_bloco() + 2] = v



from inewave.newave.modelos.blocos.versaomodelo import VersaoModelo
from inewave.newave.modelos.nwv_cortes_evap import TabelaCortesEvap

from inewave.newave.modelos.arquivoscsv.arquivocsv import ArquivoCSV
from typing import Optional
import pandas as pd  # type: ignore


class NwvCortesEvap(ArquivoCSV):
    """
    Arquivo com os cortes da evaporação linear do NEWAVE.
    """

    BLOCKS = [VersaoModelo, TabelaCortesEvap]

    @property
    def tabela(self) -> Optional[pd.DataFrame]:
        """
        A tabela de dados que está contida no arquivo.

        - periodo (`int`)
        - indice_usina (`int`)
        - nome_usina (`str`)
        - derivada_cota_area (`float`)
        - derivada_volume_cota (`float`)
        - volume_referencia_hm3 (`float`)
        - evaporacao_referencia_hm3 (`float`)
        - coeficiente_volume (`float`)
        - rhs_volume (`float`)

        :return: A tabela como um dataframe
        :rtype: pd.DataFrame | None
        """
        return self._tabela()



MockNwlistopdatOp1 = [
    " 1\n",
    "FORWARD  (ARQ. DE DADOS)    : forward.dat\n",
    "FORWARDH (ARQ. CABECALHOS)  : forwarh.dat\n",
    "NEWDESP  (REL. CONFIGS)     : newdesp.dat\n",
    "-----------------------------------------\n",
    " XXX XXX    PERIODOS INICIAL E FINAL\n",
    "   1   1\n",
    " XXXX XXXX  SERIE INICIAL E FINAL\n",
    "    1    1\n",
]

MockNwlistopdatOp2 = [
    " 2\n",
    "FORWARD  (ARQ. DE DADOS)    : forward.dat\n",
    "FORWARDH (ARQ. CABECALHOS)  : forwarh.dat\n",
    "NEWDESP  (REL. CONFIGS)     : newdesp.dat\n",
    "-----------------------------------------\n",
    " XXX XXX    PERIODOS INICIAL E FINAL\n",
    "   1  58\n",
    " 1-CMO           2-DEFICITS         3-ENA CONTROL.   4-EARM FINAL       5-ENA FIO BRUTA 6-EVAPORACAO    7-VERTIMENTO     38-SOMA AFL.PAS.                  43-VIOL. DA REST. LPP DEFL. MAX.       48-VOLUME BOMBEADO EST. BOMB.     53-VIOLACAO RHV\n",
    " 8-VAZAO MIN.    9-GER.HIDR.CONT   10-GER. TERMICA  11-INTERCAMBIOS    12-MERC.LIQ.    13-VALOR AGUA   14-VOLUME MORTO   39-GER. EOLICA                    44-RHS DA REST. LPP TURB. MAX.         49-CONSUMO EST. BOMB.             54-VALOR FORMULA RHV\n",
    "15-EXCESSO      16-GHMAX           17-OUTROS USOS   18-BENEF.INT/AGR   19-F.CORR.EC    20-GHTOTAL      21-ENA BRUTA      40-VEL. DE VENTO                  45-RHS DA REST. LPP DEFL. MAX.         50-VIOLACAO RHQ                   55-CUSTO VIOLACAO RHV\n",
    "22-ACOPLAMENTO  23-INVASAO CG      24-PENAL.INV.CG. 25-ACIONAMENTO MAR 26-COPER        27-CTERM        28-CDEFICIT       41-VIOL. DA REST. FTE             46-VIOL. DAS REST. ELETRI ESP.         51-VALOR FORMULA RHQ\n",
    "29-GER.FIO LIQ. 30-PERDA FIO       31-ENA FIO LIQ.  32-BENEF. GNL      33-VIOL.GHMIN   34-PERDAS       37-GEE            42-VIOL. DA REST. LPP TURB. MAX.  47-CUSTO VIOL. DAS REST. ELETRI ESP.   52-CUSTO VIOLACAO RHQ\n",
    " XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX (SE 99 CONSIDERA TODAS)\n",
    " 99\n",
    "-----------------------------------------------------------------------------------------------------------------------\n",
    " 1-VOL.ARMAZ       2-GER.HID         3-VOL.TURB.     4-VOL. VERT.      5-VIOL.GHMIN    6-ENCH.MORTO   7-VIOL. DEPMIN. 15 - SOMA AFL. LAG PAR(p)-A  17 - VIOL. TURB. MAX.   19 - VIOL. DA REST. LPP TURB. MAX.   21 - RHS DA REST. LPP TURB. MAX.    23 - VOLUME DESVIADO DA USINA\n",
    " 8-DESV. AGUA      9-DESV. POS.      10-DESVIO NEG.  11-VIOL. FPGHA   12-VAZAO AFL.  13-VAZAO INCREM. 14-VARM PCT.    16 - VIOL. DEFL. MAX.        18 - VIOL. TURB. MIN.   20 - VIOL. DA REST. LPP DEFL. MAX.   22 - RHS DA REST. LPP DEFL. MAX. \n",
    " XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX (SE 99 CONSIDERA TODAS)\n",
    " 99\n",
    " XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX  (SE 999 CONSIDERA TODAS AS USINAS)\n",
    " 999 \n",
]

MockNwlistopdatOp4 = [
    " 4\n",
    "FORWARD  (ARQ. DE DADOS)    : forward.dat\n",
    "FORWARDH (ARQ. CABECALHOS)  : forwarh.dat\n",
    "NEWDESP  (REL. CONFIGS)     : newdesp.dat\n",
    "-----------------------------------------\n",
    " XXX XXX    PERIODOS INICIAL E FINAL\n",
    "   1  58\n",
    " 1-CMO           2-DEFICITS         3-ENA CONTROL.   4-EARM FINAL       5-ENA FIO BRUTA 6-EVAPORACAO    7-VERTIMENTO     38-SOMA AFL.PAS.                  43-VIOL. DA REST. LPP DEFL. MAX.       48-VOLUME BOMBEADO EST. BOMB.     53-VIOLACAO RHV\n",
    " 8-VAZAO MIN.    9-GER.HIDR.CONT   10-GER. TERMICA  11-INTERCAMBIOS    12-MERC.LIQ.    13-VALOR AGUA   14-VOLUME MORTO   39-GER. EOLICA                    44-RHS DA REST. LPP TURB. MAX.         49-CONSUMO EST. BOMB.             54-VALOR FORMULA RHV\n",
    "15-EXCESSO      16-GHMAX           17-OUTROS USOS   18-BENEF.INT/AGR   19-F.CORR.EC    20-GHTOTAL      21-ENA BRUTA      40-VEL. DE VENTO                  45-RHS DA REST. LPP DEFL. MAX.         50-VIOLACAO RHQ                   55-CUSTO VIOLACAO RHV\n",
    "22-ACOPLAMENTO  23-INVASAO CG      24-PENAL.INV.CG. 25-ACIONAMENTO MAR 26-COPER        27-CTERM        28-CDEFICIT       41-VIOL. DA REST. FTE             46-VIOL. DAS REST. ELETRI ESP.         51-VALOR FORMULA RHQ\n",
    "29-GER.FIO LIQ. 30-PERDA FIO       31-ENA FIO LIQ.  32-BENEF. GNL      33-VIOL.GHMIN   34-PERDAS       37-GEE            42-VIOL. DA REST. LPP TURB. MAX.  47-CUSTO VIOL. DAS REST. ELETRI ESP.   52-CUSTO VIOLACAO RHQ\n",
    " XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX (SE 99 CONSIDERA TODAS)\n",
    " 99\n",
    "-----------------------------------------------------------------------------------------------------------------------\n",
    " 1-VOL.ARMAZ       2-GER.HID         3-VOL.TURB.     4-VOL. VERT.      5-VIOL.GHMIN    6-ENCH.MORTO   7-VIOL. DEPMIN. 15 - SOMA AFL. LAG PAR(p)-A  17 - VIOL. TURB. MAX.   19 - VIOL. DA REST. LPP TURB. MAX.   21 - RHS DA REST. LPP TURB. MAX.    23 - VOLUME DESVIADO DA USINA\n",
    " 8-DESV. AGUA      9-DESV. POS.      10-DESVIO NEG.  11-VIOL. FPGHA   12-VAZAO AFL.  13-VAZAO INCREM. 14-VARM PCT.    16 - VIOL. DEFL. MAX.        18 - VIOL. TURB. MIN.   20 - VIOL. DA REST. LPP DEFL. MAX.   22 - RHS DA REST. LPP DEFL. MAX. \n",
    " XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX XX (SE 99 CONSIDERA TODAS)\n",
    " 99\n",
    " XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX  (SE 999 CONSIDERA TODAS AS USINAS)\n",
    " 999 \n",
]



MockRegistroPostoCadastro = ("POSTO-VENTO-CAD  ; 1           ; NEInterior\n",)

MockRegistroPEEPosto = "PEE-POSTO  ; 1         ; 1\n"

MockEolicaPosto = [
    "&POSTO-VENTO-CAD ; CodigoPosto ; NomePosto\n",
    "POSTO-VENTO-CAD  ; 1           ; NEInterior\n",
    "POSTO-VENTO-CAD  ; 2           ; NELitoral\n",
    "POSTO-VENTO-CAD  ; 3           ; NEPE\n",
    "POSTO-VENTO-CAD  ; 4           ; SULInterior\n",
    "POSTO-VENTO-CAD  ; 5           ; SULLitoral\n",
    "\n",
    "&PEE-POSTO ; CodigoPEE ; CodigoPosto\n",
    "PEE-POSTO  ; 1         ; 1\n",
    "PEE-POSTO  ; 2         ; 2\n",
    "PEE-POSTO  ; 3         ; 3\n",
    "PEE-POSTO  ; 4         ; 4\n",
    "PEE-POSTO  ; 5         ; 5\n",
]



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)
from inewave.nwlistop.modelos.perdfsin import PerdfAnos


class Perdfsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao vertimento fio d'água
    , para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `perdfsin.out`.

    """

    BLOCKS = [
        PerdfAnos,
    ]



# Rotinas de testes associadas ao arquivo restricao-energia.csv do NEWAVE
from datetime import datetime
from inewave.newave.modelos.restricaoenergia import (
    RegistroRHE,
    RegistroRHEHorizPer,
    RegistroRHELsLPPEarmi,
)

from inewave.newave.restricaoenergia import RestricaoEnergia

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.restricaoenergia import (
    MockRHE,
    MockRHEHorizPer,
    MockRHQLsLPPEarmi,
    MockRestricaoEnergia,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_registro_rhe_restricaoenergia():
    m: MagicMock = mock_open(read_data="".join(MockRHE))
    r = RegistroRHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "ger_ree(1) + ener_ver_ree(1)"]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 5
    assert r.formula == "ger_ree(1) + ener_ver_ree(1)"
    r.formula = "Teste"


def test_registro_rhe_horiz_per_restricaoenergia():
    m: MagicMock = mock_open(read_data="".join(MockRHEHorizPer))
    r = RegistroRHEHorizPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2, datetime(2021, 1, 1), datetime(2025, 12, 1)]
    assert r.codigo_restricao == 2
    r.codigo_restricao = 1
    assert r.data_inicial == datetime(2021, 1, 1)
    r.data_inicial = datetime(2021, 2, 1)
    assert r.data_final == datetime(2025, 12, 1)
    r.data_final = datetime(2025, 11, 1)


def test_registro_rhe_ls_lpp_earmi_restricaoenergia():
    m: MagicMock = mock_open(read_data="".join(MockRHQLsLPPEarmi))
    r = RegistroRHELsLPPEarmi()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2, 2, -0.3, 4000.0]
    assert r.codigo_restricao == 2
    r.codigo_restricao = 1
    assert r.indice_reta == 2
    r.indice_reta = 1
    assert r.coeficiente_angular == -0.3
    r.coeficiente_angular = -0.5
    assert r.coeficiente_linear == 4000.0
    r.coeficiente_linear = 3500.0


def test_atributos_encontrados_restricaoenergia():
    m: MagicMock = mock_open(read_data="".join(MockRestricaoEnergia))
    with patch("builtins.open", m):
        e = RestricaoEnergia.read(ARQ_TESTE)
        assert len(e.rhe()) > 0
        assert len(e.rhe_horiz_per()) > 0
        assert len(e.rhe_ls_lpp_earmi()) > 0


def test_eq_restricaoenergia():
    m: MagicMock = mock_open(read_data="".join(MockRestricaoEnergia))
    with patch("builtins.open", m):
        cf1 = RestricaoEnergia.read(ARQ_TESTE)
        cf2 = RestricaoEnergia.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_restricaoenergia():
    m: MagicMock = mock_open(read_data="".join(MockRestricaoEnergia))
    with patch("builtins.open", m):
        cf1 = RestricaoEnergia.read(ARQ_TESTE)
        cf2 = RestricaoEnergia.read(ARQ_TESTE)
        cf2.data.remove(cf1.rhe()[0])
        assert cf1 != cf2


def test_leitura_escrita_restricaoenergia():
    m_leitura: MagicMock = mock_open(read_data="".join(MockRestricaoEnergia))
    with patch("builtins.open", m_leitura):
        cf1 = RestricaoEnergia.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = RestricaoEnergia.read(ARQ_TESTE)
        assert cf1 == cf2



from typing import TypeVar, List, Optional, Union
from datetime import datetime

from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.eolicaposto import (
    RegistroPEEPostoVento,
    RegistroPostoVentoCadastro,
)


class EolicaPosto(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos postos
    de vento e sua associação com os PEE.
    """

    T = TypeVar("T")

    REGISTERS = [
        RegistroPEEPostoVento,
        RegistroPostoVentoCadastro,
    ]

    def posto_vento_cad(
        self,
        codigo_posto: Optional[int] = None,
        nome_posto: Optional[str] = None,
    ) -> Optional[
        Union[
            RegistroPostoVentoCadastro,
            List[RegistroPostoVentoCadastro],
        ]
    ]:
        """
        Obtém um registro que contém o cadastro de um posto
            de vento.

        :param codigo_posto: código que identifica o posto
        :type codigo_posto: int | None
        :param nome_posto: nome do posto de vento
        :type nome_posto: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPostoVentoCadastro` |
            list[:class:`RegistroPostoVentoCadastro`] | None
        """
        return self.data.get_registers_of_type(
            RegistroPostoVentoCadastro,
            codigo_posto=codigo_posto,
            nome_posto=nome_posto,
        )

    def pee_posto(
        self,
        codigo_pee: Optional[datetime] = None,
        codigo_posto: Optional[datetime] = None,
    ) -> Optional[Union[RegistroPEEPostoVento, List[RegistroPEEPostoVento]]]:
        """
        Obtém um registro que contém o mapeamento entre um posto
            e um PEE.

        :param codigo_pee: código que identifica o PEE
        :type codigo_pee: int | None
        :param codigo_posto: código que identifica o posto
        :type codigo_posto: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroPEEPostoVento` |
            list[:class:`RegistroPEEPostoVento`] | None
        """
        return self.data.get_registers_of_type(
            RegistroPEEPostoVento,
            codigo_pee=codigo_pee,
            codigo_posto=codigo_posto,
        )



MockRHQ = "RHQ;       1 ; qtur(66)  \n"
MockRHQHorizPer = "RHQ-HORIZ-PER;        1;      2021/01;      2022/12\n"
MockRHQLsLPPVoli = "RHQ-LS-LPP-VOLI;        1;    1; 0.30; 3000.00\n"

MockRestricaoVazao = [
    "&RHQ; cod_rest;  formula\n",
    "RHQ;       1 ; qtur(66)  \n",
    "RHQ;       2 ; qtur(1) + qver(1) \n",
    "\n",
    "&RHQ-HORIZ-PER; cod_rest; data_inicial; data_final\n",
    "RHQ-HORIZ-PER;        1;      2021/01;      2022/12\n",
    "RHQ-HORIZ-PER;        2;      2021/01;      2022/12\n",
    "\n",
    "&RHQ-LS-LPP-VOLI; cod_rest; reta; coef_ang; coef_lin\n",
    "RHQ-LS-LPP-VOLI;        1;    1; 0.30; 3000.00\n",
    "RHQ-LS-LPP-VOLI;        1;    2; 0.20; 4000.00\n",
    "RHQ-LS-LPP-VOLI;        2;    1; 0.30; 1000.00\n",
    "RHQ-LS-LPP-VOLI;        2;    2; 0.20; 2000.00\n",
]
